//---------------------------------------------------------------------------
// Copyright (C) 1999 Dallas Semiconductor Corporation, All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
// OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// Except as contained in this notice, the name of Dallas Semiconductor
// shall not be used except as stated in the Dallas Semiconductor
// Branding Policy.
//---------------------------------------------------------------------------
//
//  SensorMainUn  - 1-Wire Sensor Main window using 'Tags'.
//
//  Compiler: C
//  Version: 1.00
//  Revisions:
//
#include <vcl\vcl.h>
#pragma hdrstop

#define TMEXUTIL

#include "ibtmex.h"
#include "1wsensor.h"
#include "msginc.h"
#include "SensorMainUn.h"
#include "clustun.cpp"

#define MAXTAG                100
#define SCAN_FIND_TAGS        0
#define SCAN_START_PARSE_TAG  1
#define SCAN_PARSE_TAG        2
#define SCAN_CHECK_TAGS       3

// external function prototypes
extern "C" int RunScript(Script script[], void *, int);
extern "C" void output_status(int, char *);
extern "C" int TAGToString(TAGType *, char *, int);
extern "C" int ParseTAG(TAGType *,int,int);
// globals
TAGType *TagList[MAXTAG];
int NumTags=0,VERBOSE=0,TagCount=-1;
TClusterForm *ClusterForm[MAXTAG];

// local function
void __cdecl output_status(int level, char *msg);

//---------------------------------------------------------------------------
#pragma resource "*.dfm"
TSensorMainForm *SensorMainForm;

//---------------------------------------------------------------------------
// Create event.  Do variable initialization
//
__fastcall TSensorMainForm::TSensorMainForm(TComponent* Owner)
   : TForm(Owner)
{
   // put NULL is component forms
   for (int i = 0; i < MAXTAG; i++)
      ClusterForm[i] = NULL;
}
//---------------------------------------------------------------------------
// Timer event to look for new tags and parse through the ones found.
//
void __fastcall TSensorMainForm::ScanTimerTimer(TObject *Sender)
{
   uchar tagdata[MSGMAX];
   int rslt;
   static int first=0, scan_state = SCAN_FIND_TAGS;

   // disable timer while doing step
   ScanTimer->Enabled = FALSE;

   // switch on the scan state step
   switch (scan_state)
   {
      // look for tags on the 1-Wire
      case SCAN_FIND_TAGS:
         if (RunScript(TagScan, tagdata, MSGMAX))
            AddTag(tagdata, MSGMAX);
         scan_state = SCAN_START_PARSE_TAG;
         TagCount = 0;
         break;
      // start to parse a tag
      case SCAN_START_PARSE_TAG:
         if (TagCount < NumTags)
         {
            // check if this tag should be parsed
            if (TagList[TagCount]->parse_enable)
            {
               // print tag ID and revision
               PrintTagInfo(TagList[TagCount]);

               // loop through a partial parse of the tag
               first = TRUE;
               scan_state = SCAN_PARSE_TAG;
            }
            else
               TagCount++;
         }
         else
           scan_state = SCAN_CHECK_TAGS;
         break;
      // parse a tag
      case SCAN_PARSE_TAG:
         rslt = ParseTAG(TagList[TagCount],first,FALSE);
         first = FALSE;
         // check if done with this tag
         if (rslt != NOT_DONE)
         {
            // go to next tag
            TagCount++;
            scan_state = SCAN_START_PARSE_TAG;
         }
         break;
      // check for repeat and old tags
      case SCAN_CHECK_TAGS:
         // remove old revision tags
         TagCount=-1;
         output_status(LV_ALWAYS,"");
         CheckTagRevs();
         scan_state = SCAN_FIND_TAGS;
         break;
   }

   // enable the timer again
   ScanTimer->Enabled = TRUE;
}

//---------------------------------------------------------------------------
// Form show event.  Do the setup portion of 'main'
//
void __fastcall TSensorMainForm::FormShow(TObject *Sender)
{
   // read the default PortNum and PortType
   if (!TMReadDefaultPort(&PortNum,&PortType))
   {
       Application->MessageBox("ERROR, Could not read the default PortNum and PortType!",
                    "1-Wire Sensor Application",MB_OK | MB_ICONEXCLAMATION);
       Application->Terminate();
   }

   // print the header info
   if (!PrintHeader("1-Wire Sensor GUI 1.00"))
   {
       Application->MessageBox("ERROR, Could not read the iButton-TMEX Version strings!",
                    "1-Wire Sensor Application",MB_OK | MB_ICONEXCLAMATION);
       Application->Terminate();
   }

   // run the script to setup the 1-Wire
   if (!RunScript(Verify1Wire, NULL, 0))
   {
       Application->MessageBox("1-Wire Net not successfuly setup, end program.",
                    "1-Wire Sensor Application",MB_OK | MB_ICONEXCLAMATION);
       Application->Terminate();
   }

   // check for local tag
   GetLocalTag();

   // start the scan timer
   ScanTimer->Enabled = TRUE;
}

//--------------------------------------------------------------------------
// Prints a message, and the current driver versions.  Return TRUE
// driver version were found and FALSE if not.
//
int __fastcall TSensorMainForm::PrintHeader(char *title)
{
   char tstr[80], msg[512];
   int msgcnt = 0,rt=FALSE;

   // print a header with driver version info
   sprintf(msg,"%s\n",title);
   output_status(LV_ALWAYS,msg);
   output_status(LV_ALWAYS,"Dallas Semiconductor Corporation\n");
   output_status(LV_ALWAYS,"Copyright (C) 1992-1999\n\n");
   sprintf(msg,"Port number: %d    Port type: %d\n",PortNum,PortType);
   output_status(LV_ALWAYS,msg);
   Get_Version(tstr);
   sprintf(msg,"Main driver version: %s\n",tstr);
   output_status(LV_ALWAYS,msg);
   msgcnt = sprintf(msg + msgcnt,"Sub-driver version, Type%d:",PortType);
   if (TMGetTypeVersion(PortType,tstr) < 0)
      sprintf(msg + msgcnt,"No Hardware Driver for this type found!");
   else
   {
      rt = TRUE;
      sprintf(msg + msgcnt," %s\n\n\n",tstr);
   }
   // send output
   output_status(LV_ALWAYS,msg);

   return rt;
}

//--------------------------------------------------------------------------
//  Check the current directory to see
//
int __fastcall TSensorMainForm::AddTag(uchar *tagdata, int taglen)
{
   char buf[14096];
   int i;

   // sanity check on tag data 
   if ((tagdata[0] == 0) || (tagdata[1] == 0))
      return FALSE;

   // adjust length, loop backward from end to find first non-zero
   for (i = taglen - 1; i >= 0; i--)
      if (tagdata[i] != 0)
      {
         taglen = i + 1;
         break;
      } 
 
   // create the structure to hold the Tag
   TagList[NumTags] = (TAGType *)malloc(sizeof(TAGType) + 1);

   // check if malloc was successfull 
   if (TagList[NumTags] == NULL)
   {
      output_status(LV_ALWAYS,"Insufficient memory to allocate for new Tag.\n"); 
      return FALSE;
   }
   else
      output_status(LV_VERBOSE,"Memory allocated for new Tag.\n");       

   // clear tag ID and revision
   TagList[NumTags]->tagid[0] = 0;
   TagList[NumTags]->tagrev = -1;

   // copy tag msg data to new tag
   TagList[NumTags]->msglen = taglen;
   for (i = 0; i < taglen; i++)
       TagList[NumTags]->msg[i] = tagdata[i];

   // print out the TAG Tag in human readable form
   if (TAGToString(TagList[NumTags],buf,sizeof(buf)))
   {
      // print data and status
      output_status(LV_VERBOSE,buf);
      output_status(LV_ALWAYS,"Tag is valid.\n\n");
      // mark to parse this tag
      TagList[NumTags]->parse_enable = TRUE;
      // increase the number of tags
      NumTags++;

      return TRUE;
   }
   else
   {
      // error, free memory 
      output_status(LV_ALWAYS,"Tag is not parse-able.\n\n");
      free(TagList[NumTags]);
      return FALSE;
   }      
}

//---------------------------------------------------------------------------
//  close event
//
void __fastcall TSensorMainForm::FormClose(TObject *Sender,
   TCloseAction &Action)
{
   Application->Terminate();
}

//--------------------------------------------------------------------------
//  Prints the current 'tag' information including ID number reversion if
//  available
//
void __fastcall TSensorMainForm::PrintTagInfo(TAGType *tag)
{
   char msg[512]; 
   int msgcnt = 0,i;
   time_t tlong;
   struct tm *tstruct;

   // tag ID
   if (tag->tagid[0])
   {
      msgcnt += sprintf(msg + msgcnt," Cluster ID: ");
      for (i = 1; i <= tag->tagid[0]; i++)
         msgcnt += sprintf(msg + msgcnt,"%02X",tag->tagid[i]);
   }
   // (tag revision)
   if (tag->tagrev != -1)
      msgcnt += sprintf(msg + msgcnt,"  Cluster Rev: %d ",tag->tagrev);
   // time/date 
   time(&tlong);
   tstruct = localtime(&tlong);
   sprintf(msg + msgcnt," Time: %02d/%02d/%04d  %02d:%02d:%02d\n",
       tstruct->tm_mon + 1,tstruct->tm_mday,tstruct->tm_year + 1900,   // (1.01)
       tstruct->tm_hour,tstruct->tm_min,tstruct->tm_sec);
   // send output
   output_status(LV_ALWAYS,msg);
}

//--------------------------------------------------------------------------
//  Check the current directory to see
//
int __fastcall TSensorMainForm::GetLocalTag(void)
{
   FILE *fp;
   char filename[50];
   uchar buf[MSGMAX];
   int buflen,rslt=0,i;
   char msg[255];

   // loop throuch tagd.0 to tagd.99
   for (i = 0; i < 100; i++)
   {
      // open the input file
      sprintf(filename,"TAGD.%d",i);
      fp = fopen(filename,"rb");
      if(fp == NULL)
         continue;
      else
      {
         sprintf(msg,"File '%s' opened to read.\n",filename);
         output_status(LV_ALWAYS,msg);
      }

      // read the input file
      buflen = fread(&buf[0],1,MSGMAX,fp);

      // clean up opened file
      fclose(fp);
      sprintf(msg,"File '%s' closed.\n",filename);
      output_status(LV_ALWAYS,msg);

      // add the tag to the list
      if (AddTag(buf,buflen))
         rslt = TRUE;
   }

   // check if any valid
   if (rslt)
      output_status(LV_ALWAYS,"At least one valid local TAG file found.\n");

   return rslt;
}

//--------------------------------------------------------------------------
//  Check for duplicate or older revision tags
//
void __fastcall TSensorMainForm::CheckTagRevs(void)
{
   int i,j,k,refln,tagln;

   // loop through each tag and find and compare revisions
   for (i = 0; i < NumTags; i++)
   {
      // skip if non-parsible or no ID
      if ((!TagList[i]->parse_enable) || (TagList[i]->tagid[0] == 0))
         continue;
      // extract the reference ID length
      refln = TagList[i]->tagid[0];

      // compare all of the tags
      for (j = 0; j < NumTags; j++)
      {
         // make sure not comparing same tag, non-parsible tag, or no ID
         if ((i == j) || (!TagList[j]->parse_enable) ||
                         (TagList[i]->tagid[0] == 0))
            continue;

         // get tag's ID len
         tagln = TagList[j]->tagid[0];

         // different length so not same tag
         if (refln != tagln)
            continue;

         // loop to compare
         for (k = 1; k <= tagln; k++)
            if (TagList[j]->tagid[k] != TagList[i]->tagid[k])
               break;

         // check loop
         if (k == (tagln + 1))
         {
            // same ID so now check revision
            if ((TagList[j]->tagrev == TagList[i]->tagrev) ||  // duplicate
                (TagList[j]->tagrev < TagList[i]->tagrev))     // old rev
            {
               TagList[j]->parse_enable = FALSE;
               // if cluster form visible, hide it
               if (ClusterForm[j] != NULL)
               {
                  if (ClusterForm[j]->Visible)
                     ClusterForm[j]->Hide();
               }
            }
            else if (TagList[j]->tagrev > TagList[i]->tagrev)
            {
               // newer rev so kill off the reference and break out of loop
               TagList[i]->parse_enable = FALSE;
               // if cluster form visible, hide it
               if (ClusterForm[i] != NULL)
               {
                  if (ClusterForm[i]->Visible)
                     ClusterForm[i]->Hide();
               }
               break;
            }
         }
      }
   }
}

//---------------------------------------------------------------------------
//  Output status message.  Send to main window or sub-tag windows.
//
void output_status(int level, char *msg)
{
   static int lasttag = -1;
   char tbuf[255];
   int i,cnt;
   AnsiString tans;

   // allow application to process messages
   Application->ProcessMessages();
   // release this process
   Sleep(0);

   // check for not general output
   if (TagCount != -1)
   {
      // check if current tag has a form yet
      if (ClusterForm[TagCount] == NULL)
      {
         ClusterForm[TagCount] = new TClusterForm(Application);
         sprintf(tbuf,"1-Wire Tag: #%d",TagCount);
         tans = tbuf;
         ClusterForm[TagCount]->Caption = tans;
         ClusterForm[TagCount]->Show();
      }
   }
   // check if this is a new output stream
   if (lasttag != TagCount)
   {
      // close the last one
      if (lasttag != -1)
      {
         ClusterForm[lasttag]->EndOutput();
         if (!ClusterForm[lasttag]->Visible)
            ClusterForm[lasttag]->Show();
      }
      if (TagCount != -1)
         // start the next one
         ClusterForm[TagCount]->StartOutput();
   }
   lasttag = TagCount;

   // check for null string
   if (msg[0] == 0)
      return;

   // see if just status
   if (level != LV_ALWAYS)
   {
      if (TagCount == -1)
         SensorMainForm->SetStatus(msg);
      else
         ClusterForm[TagCount]->SetStatus(msg);
      return;
   }

   tbuf[0] = 0;
   cnt = 0;
   for (i = 0; i < (int)strlen(msg); i++)
   {
      if ((msg[i] == 0x0A) || (i == ((int)strlen(msg)-1)) || (cnt > 250))
      {
         tbuf[cnt] = 0;
         tans = tbuf;
         if (TagCount == -1)
            SensorMainForm->ShowMemo->Lines->Add(tans);
         else
            ClusterForm[TagCount]->AddOutput(tans);
         cnt = 0;
      }
      else
         tbuf[cnt++] = msg[i];
   }
}

//---------------------------------------------------------------------------
// Exit menu event
//
void __fastcall TSensorMainForm::ExitMenuClick(TObject *Sender)
{
   Application->Terminate();
}

//---------------------------------------------------------------------------
// About event
//
void __fastcall TSensorMainForm::AboutMenuClick(TObject *Sender)
{
  Application->MessageBox("1-Wire Sensor Application, Version 1.00",
                    "1-Wire Sensor Application",MB_OK);
}

//---------------------------------------------------------------------------
// Display status
//
void __fastcall TSensorMainForm::SetStatus(char *msg)
{
   if (!((msg[0] == '.') && (msg[1] == 0)))
      StatusBar->SetTextBuf(msg);
}

//---------------------------------------------------------------------------
